/*
  Cell shader for nuiMesh
  Copyright (C) 2002 Vincent Caron <v.caron@zerodeux.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "CellEngine.h"

using namespace std;


CellEngine::CellEngine (nglContext* pContext)
  : nuiMeshEngine(pContext)
{
  mVertexProgram = -1;
  mFragmentProgram = -1;

  if (!mpContext)
  {
    NGL_OUT("CellEngine: GL context n/a\n");
    return;
  }
  if (!mpContext->CheckExtension(_T("GL_ARB_multitexture")) ||
      !mpContext->CheckExtension(_T("GL_EXT_vertex_shader")) ||
      !mpContext->CheckExtension(_T("GL_ATI_fragment_shader")))
  {
    NGL_OUT("CellEngine: required GL extensions not found\n");
    return;
  }

  Setup();
}

CellEngine::~CellEngine()
{
  if (!mpContext)
    return;

  if (mVertexProgram != -1)
    mpContext->glDeleteVertexShaderEXT(mVertexProgram);

  if (mFragmentProgram != -1)
    mpContext->glDeleteFragmentShaderATI(mFragmentProgram);
}



void CellEngine::SetCellLight (uint Index)
{
  mLightIndex = Index;
}

#define GAMMA(x, gamma) (pow((x), (gamma)))

void CellEngine::UpdateLightRamp()
{
  uint i, s;

  uint shade_cnt = (uint)mLightShadeCnt;
  shade_cnt = MAX(1, shade_cnt);

  uint ambient = (uint)(mLightAmbient * 256.f);

  uint i_begin;
  uint i_end = 0;

  for (s = 0; s < shade_cnt; s++)
  {
    i_begin = i_end;
    i_end  = (uint)(GAMMA((float)(s + 1) / (float)shade_cnt, mLightGammaX) * 256.f);

    float y0 = (float)GAMMA(((float)s)       / (float)shade_cnt, mLightGammaY);
    float y1 = (float)GAMMA(((float)s + 1.f) / (float)shade_cnt, mLightGammaY);

    float dx = (float)(i_end - i_begin);
    float dy = (y1 - y0) * mLightFlatness;

    for (i = i_begin; i < i_end; i++)
    {
      float x = (float)(i - i_begin) / dx;
      uint y = (uint8)((y1 - (1.f - x) * dy) * 255.f);

      mLightRamp[i] = MAX(ambient, y);
    }
  }
}

void CellEngine::UpdateEdgeRamp()
{
  uint i;

  float thick = mEdgeThickness * EDGERAMP_SIZE;
  uint i_thick = (uint)thick;
  uint i_top1  = (uint)(mEdgeFalloff * thick);
  uint i_top2  = (uint)((1.f - mEdgeSharpness) * thick);

  for (i = 0; i < i_top1; i++)
    mEdgeRamp[i] = (GLubyte)(255.f * (mEdgeFalloffOffset + (mEdgeColor[3] - mEdgeFalloffOffset) * (float)i / (float)i_top1));

  for (i = i_top1; i < i_top2; i++)
    mEdgeRamp[i] = (GLubyte)(255.f * mEdgeColor[3]);

  for (i = i_top2; i < i_thick; i++)
    mEdgeRamp[i] = (GLubyte)(255.f * mEdgeColor[3] * (float)(i_thick - i) / (float)(i_thick - i_top2));

  for (i = i_thick; i < EDGERAMP_SIZE; i++)
    mEdgeRamp[i] = 0;
}


/* Texture units usage
 */
#define NORMAL_TEXU     0
#define EYE_TEXU        1
#define LIGHT_TEXU      2
#define EDGERAMP_TEXU   3
#define LIGHTRAMP_TEXU  4


void CellEngine::SetupTexture (GLenum Unit, GLenum Target)
{
  mpContext->glActiveTextureARB(GL_TEXTURE0_ARB + Unit);

  // Target == GL_NONE disable all targets
  if (Target == GL_TEXTURE_1D) glEnable (GL_TEXTURE_1D); else glDisable(GL_TEXTURE_1D);
  if (Target == GL_TEXTURE_2D) glEnable (GL_TEXTURE_2D); else glDisable(GL_TEXTURE_2D);
  if (Target == GL_TEXTURE_3D) glEnable (GL_TEXTURE_3D); else glDisable(GL_TEXTURE_3D);

  if (Target == GL_NONE)
    return;

  if (Target >= GL_TEXTURE_1D) glTexParameterf(Target, GL_TEXTURE_WRAP_S, GL_CLAMP);
  if (Target >= GL_TEXTURE_2D) glTexParameterf(Target, GL_TEXTURE_WRAP_T, GL_CLAMP);
  if (Target >= GL_TEXTURE_3D) glTexParameterf(Target, GL_TEXTURE_WRAP_R, GL_CLAMP);
  /* If we use linear interpolation below, border pixels (s=0 & s=1) are averaged
   * with the texture border color (black as a default). We want an exact clamping
   * to tex[0] and tex[1], so we go GL_NEAREST.
   */
  glTexParameterf(Target, GL_TEXTURE_MAG_FILTER, /*GL_LINEAR*/GL_NEAREST);
  glTexParameterf(Target, GL_TEXTURE_MIN_FILTER, /*GL_LINEAR*/GL_NEAREST);
}

void CellEngine::Draw (nuiDrawContext* pContext, const nuiRect& rViewPort)
{
  // Force a white background
  ngl3DSColor& bg = mScene.mpBgColor[0];
  bg.mR = 1.f;
  bg.mG = 1.f;
  bg.mB = 1.f;

  /* Setup cellengine context
   */
  glPushAttrib(GL_TEXTURE_BIT);
  SetupTexture(NORMAL_TEXU, GL_TEXTURE_3D);
  SetupTexture(EYE_TEXU, GL_TEXTURE_3D);
  SetupTexture(LIGHT_TEXU, GL_TEXTURE_3D);
  SetupTexture(EDGERAMP_TEXU, GL_TEXTURE_1D);
  glTexImage1D(GL_TEXTURE_1D, 0, GL_LUMINANCE8, 256, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, mEdgeRamp);
  SetupTexture(LIGHTRAMP_TEXU, GL_TEXTURE_1D);
  glTexImage1D(GL_TEXTURE_1D, 0, GL_LUMINANCE8, 256, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, mLightRamp);

  glEnable(GL_VERTEX_SHADER_EXT);
  glEnable(GL_FRAGMENT_SHADER_ATI);

  /* Draw content
   */
  nuiMeshEngine::Draw(pContext, rViewPort);

  /* Restore context
   */
  SetupTexture(NORMAL_TEXU,     GL_NONE);
  SetupTexture(EYE_TEXU,        GL_NONE);
  SetupTexture(LIGHT_TEXU,      GL_NONE);
  SetupTexture(EDGERAMP_TEXU,   GL_NONE);
  SetupTexture(LIGHTRAMP_TEXU,  GL_NONE);
  glPopAttrib();

  glDisable(GL_VERTEX_SHADER_EXT);
  glDisable(GL_FRAGMENT_SHADER_ATI);
}

void CellEngine::OnDraw (nuiDrawContext* pContext)
{
  nglMatrixf mv;
  mv.GetMatrix(GL_MODELVIEW_MATRIX);

  /* Vertex program setup (all vertices share the same program)
   */
  mpContext->glBindVertexShaderEXT(mVertexProgram);

  ngl3DSLight& light = mLight[mLightIndex];
  nglVectorf lpos(light.mPos.mX, light.mPos.mY, light.mPos.mZ);
  lpos = mv * lpos; // Vertex program operates in eye space
  mpContext->glSetInvariantEXT(mLightPosVar, GL_FLOAT, lpos.Elt);


  /* Fragment program setup (all polygons share the same program)
   */
  mpContext->glBindFragmentShaderATI(mFragmentProgram);
  mpContext->glSetFragmentShaderConstantATI(GL_CON_1_ATI, mEdgeColor);

  nuiMeshEngine::OnDraw(pContext);
}

void CellEngine::SetupMaterial(uint Index)
{
  /* We completely override and ignore nuiMeshEngine::SetupMaterial */
  ngl3DSMaterial& mat = mMaterial[Index];
  GLfloat mat_color[] = { mat.mDiffuse.mR, mat.mDiffuse.mG, mat.mDiffuse.mB, 1.f};

  mpContext->glSetFragmentShaderConstantATI(GL_CON_0_ATI, mat_color);
}



void CellEngine::Setup()
{
  mEdgeColor[0] = 0.f;
  mEdgeColor[1] = 0.f;
  mEdgeColor[2] = 0.f;
  mEdgeColor[3] = 1.f;
  mEdgeThickness = .2f;
  mEdgeSharpness = .25f;
  mEdgeFalloff   = .15f;
  mEdgeFalloffOffset = .3f;
  UpdateEdgeRamp();

  mLightIndex = 0;
  mLightShadeCnt = 1;
  mLightFlatness = 1.f;
  mLightGammaX = 1.f;
  mLightGammaY = 1.f;
  mLightAmbient = .0f;
  UpdateLightRamp();

  SetupVertexProgram();
  SetupFragmentProgram();
}

void CellEngine::SetupVertexProgram()
{
  mLightPosVar  = mpContext->glGenSymbolsEXT(GL_VECTOR_EXT, GL_INVARIANT_EXT, GL_FULL_RANGE_EXT, 1);

  /* Bind some GL states to vertex shader parameters */
  GLuint modelview  = mpContext->glBindParameterEXT(GL_MODELVIEW_MATRIX);
  GLuint projection = mpContext->glBindParameterEXT(GL_PROJECTION_MATRIX);
  GLuint vertex     = mpContext->glBindParameterEXT(GL_CURRENT_VERTEX_EXT);
  GLuint normal     = mpContext->glBindParameterEXT(GL_CURRENT_NORMAL);

  mVertexProgram = mpContext->glGenVertexShadersEXT(1);
  mpContext->glBindVertexShaderEXT(mVertexProgram);

  mpContext->glBeginVertexShaderEXT();
  {
    GLuint vertex_t = mpContext->glGenSymbolsEXT(GL_VECTOR_EXT, GL_LOCAL_EXT, GL_FULL_RANGE_EXT, 1);
    GLuint vector   = mpContext->glGenSymbolsEXT(GL_VECTOR_EXT, GL_LOCAL_EXT, GL_FULL_RANGE_EXT, 1);
    GLuint scalar   = mpContext->glGenSymbolsEXT(GL_SCALAR_EXT, GL_LOCAL_EXT, GL_FULL_RANGE_EXT, 1);

    /* Transform and project vertex.
     * This is done in two steps (instead of single multiply by MVP) because we need
     * the intermediate transformed (and _not_ projected) vertex_t value.
     */
    mpContext->glShaderOp2EXT(GL_OP_MULTIPLY_MATRIX_EXT, vertex_t, modelview, vertex);
    mpContext->glShaderOp2EXT(GL_OP_MULTIPLY_MATRIX_EXT, GL_OUTPUT_VERTEX_EXT, projection, vertex_t);


    /* Output normal to the NORMAL_TEXU texture unit
     */
    mpContext->glShaderOp2EXT(GL_OP_MULTIPLY_MATRIX_EXT, GL_OUTPUT_TEXTURE_COORD0_EXT + NORMAL_TEXU, modelview, normal);

    /* Output normalized eye vector to the EYE_TEXU texture unit
     * The eye vector is simply -vertex_t since the camera model brought our eye to the origin.
     */
    mpContext->glSwizzleEXT(vector, vertex_t, GL_NEGATIVE_X_EXT, GL_NEGATIVE_Y_EXT, GL_NEGATIVE_Z_EXT, GL_ZERO_EXT);
    mpContext->glShaderOp2EXT(GL_OP_DOT3_EXT, scalar, vector, vector);
    mpContext->glShaderOp1EXT(GL_OP_RECIP_SQRT_EXT, scalar, scalar);
    mpContext->glShaderOp2EXT(GL_OP_MUL_EXT, GL_OUTPUT_TEXTURE_COORD0_EXT + EYE_TEXU, vector, scalar);

    /* Output normalized light vector to LIGHT_TEXU texture unit
     */
    mpContext->glShaderOp2EXT(GL_OP_SUB_EXT, vector, mLightPosVar, vertex_t);
    mpContext->glShaderOp2EXT(GL_OP_DOT3_EXT, scalar, vector, vector);
    mpContext->glShaderOp1EXT(GL_OP_RECIP_SQRT_EXT, scalar, scalar);
    mpContext->glShaderOp2EXT(GL_OP_MUL_EXT, GL_OUTPUT_TEXTURE_COORD0_EXT + LIGHT_TEXU, vector, scalar);
  }
  mpContext->glEndVertexShaderEXT();

  if (glGetError() != GL_NO_ERROR)
    NGL_OUT("CellEngine: vertex program setup failed\n");
}

/* Constant mapping :
 *
 *  CON_0 : material color
 *  CON_1 : edge color
 */
void CellEngine::SetupFragmentProgram()
{
  mFragmentProgram = mpContext->glGenFragmentShadersATI(1);
  mpContext->glBindFragmentShaderATI(mFragmentProgram);

  mpContext->glBeginFragmentShaderATI();
  {
    /* 1st pass
     */
    mpContext->glPassTexCoordATI(GL_REG_0_ATI + NORMAL_TEXU, GL_TEXTURE0_ARB + NORMAL_TEXU, GL_SWIZZLE_STR_ATI);
    mpContext->glPassTexCoordATI(GL_REG_0_ATI + EYE_TEXU,    GL_TEXTURE0_ARB + EYE_TEXU,    GL_SWIZZLE_STR_ATI);
    mpContext->glPassTexCoordATI(GL_REG_0_ATI + LIGHT_TEXU,  GL_TEXTURE0_ARB + LIGHT_TEXU,  GL_SWIZZLE_STR_ATI);

    // Silhouette factor (N.E)
    mpContext->glColorFragmentOp2ATI(
      GL_DOT3_ATI,
      GL_REG_4_ATI,               GL_NONE, GL_NONE,   /* r4 = N.E */
      GL_REG_0_ATI + NORMAL_TEXU, GL_NONE, GL_NONE,   /* N */
      GL_REG_0_ATI + EYE_TEXU,    GL_NONE, GL_NONE);  /* E */

    // Light Lambert coef (N.L), clamped to [0, 1]
    mpContext->glColorFragmentOp2ATI(
      GL_DOT3_ATI,
      GL_REG_5_ATI,               GL_NONE, GL_SATURATE_BIT_ATI,   /* r5 = N.L */
      GL_REG_0_ATI + NORMAL_TEXU, GL_NONE, GL_NONE,   /* N */
      GL_REG_0_ATI + LIGHT_TEXU,  GL_NONE, GL_NONE);  /* L */


    /* 2nd pass
     */
    mpContext->glSampleMapATI(GL_REG_0_ATI + EDGERAMP_TEXU,  GL_REG_4_ATI, GL_SWIZZLE_STR_ATI);  /* r[sil] = silhouette_ramp[r4] */
    mpContext->glSampleMapATI(GL_REG_0_ATI + LIGHTRAMP_TEXU, GL_REG_5_ATI, GL_SWIZZLE_STR_ATI);  /* r[light] = light_ramp[r5] */

    // r0 = lighted material (material color * light Lambert coef)
    mpContext->glColorFragmentOp2ATI(
      GL_MUL_ATI,
      GL_REG_0_ATI, GL_NONE, GL_NONE,
      GL_CON_0_ATI, GL_NONE, GL_NONE,
      GL_REG_0_ATI + LIGHTRAMP_TEXU, GL_NONE, GL_NONE);

   // r1 = silhouette alpha
    mpContext->glColorFragmentOp1ATI(
      GL_MOV_ATI,
      GL_REG_1_ATI, GL_NONE, GL_NONE,
      GL_REG_0_ATI + EDGERAMP_TEXU, GL_NONE, GL_NONE);

    // Blend the silhouette color (CON1) with the lighted material (r0), using alpha factor r1
    mpContext->glColorFragmentOp3ATI(
      GL_LERP_ATI,
      GL_REG_0_ATI, GL_NONE, GL_NONE,  // result
      GL_REG_1_ATI, GL_NONE, GL_NONE,  // alpha
      GL_CON_1_ATI, GL_NONE, GL_NONE,  // color
      GL_REG_0_ATI, GL_NONE, GL_NONE); // 1 - color
  }
  mpContext->glEndFragmentShaderATI();

  if (glGetError() != GL_NO_ERROR)
    NGL_OUT("CellEngine: fragment program setup failed\n");
}



void CellEngine::OnElement (const ngl3DSMesh& rMesh, const ElementType& rElement)
{
/*
  ngl3DSMaterial& mat = mMaterial[rElement.mMaterial];

  GLfloat mat_color[] = { mat.mDiffuse.mR, mat.mDiffuse.mG, mat.mDiffuse.mB, 1.f};

  mpContext->glVariantfvEXT(mMatColorVar, mat_color);
*/
}

bool CellEngine::OnMesh (const ngl3DSMesh& rMesh)
{
  if (!nuiMeshEngine::OnMesh(rMesh))
    return false;

#if 0
  ElementType elt;
  uint f, v;

  /* Record the wire frame version of the element
   */
  elt.mList = glGenLists(1);
  elt.mMaterial = 0; // Not used

  glNewList(elt.mList, GL_COMPILE);
  {
    for (f = 0; f < rMesh.mFaceCnt; f++)
    {
      ngl3DSFace& face = rMesh.mpFace[f];

      glBegin(GL_LINE_LOOP);

      for (v = 0; v < 3; v++)
      {
        uint16        vertex_id = face.mpVertex[v];
        ngl3DSVertex& vertex    = rMesh.mpVertex[vertex_id];

        glVertex3fv(&vertex.mX);
      }

      glEnd();
    }
  }
  glEndList();

  mLineElement.push_back(elt);
#endif
  return true;
}
