/*
  Cell shader for nuiMesh
  Copyright (C) 2002 Vincent Caron <v.caron@zerodeux.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "nuiApplication.h"
#include "nuiMainWindow.h"
#include "CellMesh.h"

using namespace std;


#define MARGIN 5

/*
 * Life cycle
 */

CellMesh::CellMesh (nuiContainerPtr pParent) :
  nuiMesh(pParent)
{
  mEventSink.SetTarget(this);

  nuiMainWindow* main = dynamic_cast<nuiMainWindow*>(GetRoot());

  mpFileDialog = NULL;

  mPanMode = ePanCamera;
  mCellLight = 0;
  mpCellEngine = new CellEngine(main);
  SetEngine(mpCellEngine);

  /* Setup ramp editors
   */

  nuiRect pos((nuiSize)50, (nuiSize)50, (nuiSize)(512 + MARGIN*2), (nuiSize)(256 + MARGIN*2));
  mpEdgeEdit = new RampEdit(main, pos, mpCellEngine, RampEdit::eEdge);
  mpEdgeEdit->SetVisible(false);
  mpEdgeEdit->AddControl(nglString(_T("Thickness")),      &mpCellEngine->mEdgeThickness);
  mpEdgeEdit->AddControl(nglString(_T("Sharpness")),      &mpCellEngine->mEdgeSharpness);
  mpEdgeEdit->AddControl(nglString(_T("Falloff")),        &mpCellEngine->mEdgeFalloff);
  mpEdgeEdit->AddControl(nglString(_T("Falloff offset")), &mpCellEngine->mEdgeFalloffOffset);
  mpEdgeEdit->AddControl(nglString(_T("Red")),            &mpCellEngine->mEdgeColor[0]);
  mpEdgeEdit->AddControl(nglString(_T("Green")),          &mpCellEngine->mEdgeColor[1]);
  mpEdgeEdit->AddControl(nglString(_T("Blue")),           &mpCellEngine->mEdgeColor[2]);
  mpEdgeEdit->AddControl(nglString(_T("Alpha")),          &mpCellEngine->mEdgeColor[3]);

  pos.Move(50, 50);
  mpLightEdit = new RampEdit(main, pos, mpCellEngine, RampEdit::eLight);
  mpLightEdit->SetVisible(false);
  mpLightEdit->AddControl(nglString(_T("Shades")),         &mpCellEngine->mLightShadeCnt, 0.f, 15.f);
  mpLightEdit->AddControl(nglString(_T("Flatness")),       &mpCellEngine->mLightFlatness);
  mpLightEdit->AddControl(nglString(_T("Gamma X")),        &mpCellEngine->mLightGammaX, .1f, 5.f);
  mpLightEdit->AddControl(nglString(_T("Gamma Y")),        &mpCellEngine->mLightGammaY, .1f, 5.f);
  mpLightEdit->AddControl(nglString(_T("Ambient")),        &mpCellEngine->mLightAmbient);
}

CellMesh::~CellMesh()
{
  if (mpEdgeEdit)  mpEdgeEdit->Trash();
  if (mpLightEdit) mpLightEdit->Trash();
}

void CellMesh::Load (const nglPath& rPath)
{
  nuiXML data(nglString("cellshader"));

  data.Load(nglIFile(rPath));
  if (data.GetName() == "cellshader")
  {
    mpCellEngine->mEdgeColor[0] = ReadValue (&data, "edge/color/red");
    mpCellEngine->mEdgeColor[1] = ReadValue (&data, "edge/color/green");
    mpCellEngine->mEdgeColor[2] = ReadValue (&data, "edge/color/blue");
    mpCellEngine->mEdgeColor[3] = ReadValue (&data, "edge/color/alpha");
    mpCellEngine->mEdgeThickness = ReadValue (&data, "edge/thickness");
    mpCellEngine->mEdgeSharpness = ReadValue (&data, "edge/sharpness");
    mpCellEngine->mEdgeFalloff = ReadValue (&data, "edge/falloff");
    mpCellEngine->mEdgeFalloffOffset = ReadValue (&data, "edge/falloff_offset");
    mpCellEngine->UpdateEdgeRamp();

    mpCellEngine->mLightShadeCnt = ReadValue (&data, "light/shades");
    mpCellEngine->mLightFlatness = ReadValue (&data, "light/flatness");
    mpCellEngine->mLightGammaX = ReadValue (&data, "light/gamma_x");
    mpCellEngine->mLightGammaY = ReadValue (&data, "light/gamma_y");
    mpCellEngine->mLightAmbient = ReadValue (&data, "light/ambient");
    mpCellEngine->UpdateLightRamp();
  }
  else
    NGL_OUT("Error: '%s' is not a Cell Shader description file\n", rPath.GetChars());
}

void CellMesh::Save (const nglPath& rPath)
{
  nglString s;
  nglString data = "<?xml version=\"1.0\"?>\n";

  data += "<cellshader>\n";

  data += " <edge>\n";
  data += "  <color>\n";
  data += "   " + FormatValue("red",   mpCellEngine->mEdgeColor[0]);
  data += "   " + FormatValue("green", mpCellEngine->mEdgeColor[1]);
  data += "   " + FormatValue("blue",  mpCellEngine->mEdgeColor[2]);
  data += "   " + FormatValue("alpha", mpCellEngine->mEdgeColor[3]);
  data += "  </color>\n";
  data += "  " + FormatValue("thickness", mpCellEngine->mEdgeThickness);
  data += "  " + FormatValue("sharpness", mpCellEngine->mEdgeSharpness);
  data += "  " + FormatValue("falloff", mpCellEngine->mEdgeFalloff);
  data += "  " + FormatValue("falloff_offset", mpCellEngine->mEdgeFalloffOffset);
  data += " </edge>\n";

  data += " <light>\n";
  data += "  " + FormatValue("shades", mpCellEngine->mLightShadeCnt);
  data += "  " + FormatValue("flatness", mpCellEngine->mLightFlatness);
  data += "  " + FormatValue("gamma_x", mpCellEngine->mLightGammaX);
  data += "  " + FormatValue("gamma_y", mpCellEngine->mLightGammaY);
  data += "  " + FormatValue("ambient", mpCellEngine->mLightAmbient);
  data += " </light>\n";

  data += "</cellshader>\n";

  nglOFile file(rPath, eOFileCreate);
  file.WriteText(data);
}


/*
 * Navigation
 */

bool CellMesh::KeyDown (nglKeyCode Key, nglChar Char)
{
  if (Key == NK_ESC)
  {
    App->Quit();
    return true;
  }

  bool update = true;

  switch (Char)
  {
    case _T('e'):
    case _T('E'):
      mpEdgeEdit->SetVisible(!mpEdgeEdit->IsVisible());
      break;

    case _T('l'):
    case _T('L'):
      mpLightEdit->SetVisible(!mpLightEdit->IsVisible());
      break;

    case _T('n'):
    case _T('N'):
      mCellLight = (mCellLight + 1) % mpCellEngine->GetLightCnt();
      mpCellEngine->SetCellLight(mCellLight);
      NGL_OUT("CellMesh: selecting cell light #%d (/%d)\n", mCellLight, mpCellEngine->GetLightCnt());
      break;

    case _T('p'):
    case _T('P'):
      switch (mPanMode)
      {
        case ePanCamera : mPanMode = ePanLight; NGL_OUT(_T("Panning: light mode\n")); break;
        case ePanLight  : mPanMode = ePanCamera; NGL_OUT(_T("Panning: camera mode\n"));  break;
      }
      break;

    case _T('o'):
    case _T('O'):
      if (!mpFileDialog)
      {
        nglPath path(ePathCurrent);
        mpFileDialog = new nuiFileSelector(GetRoot(), path, true);
        mpFileDialog->SetTitle(nglString("Open a shader preset"));
        mEventSink.Connect(mpFileDialog->Done, &CellMesh::OnFileLoad);
      }
      break;

    case _T('s'):
    case _T('S'):
      if (!mpFileDialog)
      {
        nglPath path(ePathCurrent);
        mpFileDialog = new nuiFileSelector(GetRoot(), path, true);
        mpFileDialog->SetTitle(nglString("Save current shader preset as"));
        mEventSink.Connect(mpFileDialog->Done, &CellMesh::OnFileSave);
      }
      break;

    default:
      update = false;
      if (nuiMesh::KeyDown(Key, Char))
        return true;
  }

  if (update)
  {
    Invalidate();
    return true;
  }

  return false;
}

nglString CellMesh::FormatValue (const nglChar* pName, float Value)
{
  nglString temp;

  temp.CFormat("<%s>%f</%s>\n", pName, Value, pName);
  return temp;
}

bool CellMesh::OnFileSave (nuiEvent* pEvent)
{
  nglPath path;

  if (mpFileDialog->GetSelection(path))
    Save(path);
  mpFileDialog = NULL;

  Focus();
  return true;
}

float CellMesh::ReadValue (nuiXMLNode* pNode, const nglChar* pPath)
{
  nuiXMLNode* node;

  node = pNode->Find(nglString(pPath));
  if (!node)
  {
    NGL_OUT("Warning: %s value not found, defaulting to 0\n", pPath);
    return 0.f;
  }

  nglString val = node->GetChild(0)->GetValue();
  return val.GetCFloat();
}

bool CellMesh::OnFileLoad (nuiEvent* pEvent)
{
  nglPath path;

  if (mpFileDialog->GetSelection(path))
    Load(path);
  mpFileDialog = NULL;

  Focus();
  return true;
};

void CellMesh::StartView (nuiSize X, nuiSize Y)
{
  nuiMesh::StartView(X, Y);

  if (mPanMode == ePanLight)
  {
    /* Save current light positions
     */
    uint i;
    mLightPosRef.clear();

    for (i = 0; i < mpCellEngine->GetLightCnt(); i++)
    {
      nglVectorf vec;

      mpCellEngine->GetLight(i, &vec, NULL);
      mLightPosRef.push_back(vec);
    }
  }
}

bool CellMesh::UpdateView (float DeltaX, float DeltaY)
{
  if (mViewMode == eViewNone)
    return false;

  switch (mPanMode)
  {
    case ePanCamera: return nuiMesh::UpdateView(DeltaX, DeltaY);

    case ePanLight:
    {
      nglMatrixf rot_x, rot_y;
      uint i;

      rot_x.SetRotation(DeltaY, mHorizon);
      rot_y.SetRotation(-DeltaX, nglVectorf(0.f, 1.f, 0.f));

      for (i = 0; i < mLightPosRef.size(); i++)
      {
        nglVectorf pos = mLightPosRef[i];
        pos = rot_x * pos;
        pos = rot_y * pos;
        mpCellEngine->SetLight(i, &pos, NULL);
      }
      Invalidate();
    }
  }

  return true;
}
