/*
  Cell shader for nuiMesh
  Copyright (C) 2002 Vincent Caron <v.caron@zerodeux.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "nuiApplication.h"
#include "nuiLabel.h"
#include "nuiScrollBar.h"
#include "RampEdit.h"

using namespace std;


/*
 * RampView
 */

class RampView : public nuiUserArea
{
public:
  RampView(nuiContainer* pParent, uint8* pData, uint ByteCount);
  virtual ~RampView() {};

protected:
  bool OnUserDraw (nuiEvent* pEvent);

private:
  nuiEventSink<RampView> mUserDrawSink;

  uint8* mpRampData;
  uint    mRampSize;
};


RampView::RampView (nuiContainer* pParent, uint8* pData, uint ByteCount)
 : nuiUserArea(pParent)
{
  mUserDrawSink.SetTarget(this);
  mUserDrawSink.Connect(this->UserDraw, &RampView::OnUserDraw, GetDrawContext());

  mpRampData = pData;
  mRampSize = ByteCount;
}

bool RampView::OnUserDraw (nuiEvent* pEvent)
{
  if (!mpRampData)
    return false;

/*
  nuiDrawContext* pContext = (nuiDrawContext*)pEvent->mpUser;
  pContext->PushClipping();
  pContext->SetClipping(true);
  pContext->Clip(this);

  mpMesh->Draw(pContext, mRect);

  pContext->PopClipping();
*/
  nuiRect rect = GetRect();
  float height = rect.GetHeight();
  float x_scale = (float)rect.GetWidth() / (float)mRampSize;
  float y_scale = height / 255.f;

  uint i;

  glBegin(GL_LINE_STRIP);
  {
    glColor4f(0.f, 0.f, 0.f, 1.f);

    for (i = 0; i < mRampSize; i++)
      glVertex2f((float)i * x_scale, height - (float)mpRampData[i] * y_scale);
  }
  glEnd();

  return true;
}



/*
 * RampControl
 */

class RampControl : public nuiScrollBar
{
public:
  RampControl(nuiContainer* pParent, float* pValue, float Min, float Max, RampEdit* pRampEdit);
  virtual ~RampControl() {};

  void SetLabel(nuiLabel* pLabel);

private:
  nuiEventSink<RampControl> mValueChangedSink;

  RampEdit* mpRampEdit;
  float*    mpValue;
  nuiLabel* mpLabel;

  bool Draw (nuiDrawContext* pContext);
  void UpdateLabel();
  bool OnValueChanged(nuiEvent* pEvent);
};

RampControl::RampControl (nuiContainer* pParent, float* pValue, float Min, float Max, RampEdit* pRampEdit)
 : nuiScrollBar(pParent, nuiHorizontal, nuiRange(pValue ? *pValue : 0.f, Min, Max + (Max - Min) / 10.f, .01f, .1f, (Max - Min) / 10.f))
{
  mValueChangedSink.SetTarget(this);
  mValueChangedSink.Connect(this->ValueChanged, &RampControl::OnValueChanged);

  mpValue = pValue;
  mpRampEdit = pRampEdit;
  mpLabel = NULL;
}

void RampControl::SetLabel (nuiLabel* pLabel)
{
  mpLabel = pLabel;
  UpdateLabel();
}

bool RampControl::Draw (nuiDrawContext* pContext)
{
  float val = mRange.GetValue();

  if (mpValue && *mpValue != val)
    mRange.SetValue(*mpValue);

  return nuiScrollBar::Draw(pContext);
}

void RampControl::UpdateLabel()
{
  if (mpLabel && mpValue)
  {
    nglString txt;
    txt.Format(_T("%.3f"), *mpValue);
    mpLabel->SetText(txt);
  }
}

bool RampControl::OnValueChanged(nuiEvent* pEvent)
{
  float val = mRange.GetValue();

  if (mpValue)
    *mpValue = val;

  UpdateLabel();
  mpRampEdit->OnValueChanged();

  return true;
}



/*
 * RampEdit
 */

RampEdit::RampEdit (nuiMainWindow* pParent, const nuiRect& rRect,  CellEngine* pEngine, RampType Ramp)
 : nuiWindow(pParent,
             rRect,
             nuiWindow::NoResize | nuiWindow::NoClose)
{
  uint8* ramp_data;
  uint ramp_size;

  switch (Ramp)
  {
    case eEdge:
      SetTitle(nglString(_T("Edge settings")));
      ramp_data = pEngine->mEdgeRamp;
      ramp_size = EDGERAMP_SIZE;
      break;

    case eLight:
      SetTitle(nglString(_T("Light settings")));
      ramp_data = pEngine->mLightRamp;
      ramp_size = LIGHTRAMP_SIZE;
      break;
  }

  mpEngine = pEngine;

  // First horizontal box
  mpMainBox = new nuiWidgetBox(this, nuiHorizontal);
  mpMainBox->SetProportional(false);

  // Left pane
  mpRampView = new RampView(mpMainBox, ramp_data, ramp_size);

  // Right pane, controls holder
  mpCtlBox = new nuiWidgetBox(mpMainBox, nuiVertical);
  mpCtlBox->SetFill(false);
  mpCtlBox->SetProportional(false);
}

RampEdit::~RampEdit()
{
}


void RampEdit::AddControl (const nglString& rTitle, float* pValue, float Min, float Max)
{
  nuiWidgetBox* box = new nuiWidgetBox(mpCtlBox, nuiHorizontal);
  box->SetFill(true);
  box->SetProportional(false);

  new nuiLabel(box, rTitle);
  RampControl* ctl = new RampControl(box, pValue, Min, Max, this);
  nuiLabel* label  = new nuiLabel(box, nglString::Empty);
  ctl->SetLabel(label);
}

void RampEdit::OnValueChanged()
{
  mpEngine->UpdateEdgeRamp();
  mpEngine->UpdateLightRamp();
}
